import os
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

class Watchdog:
  def __init__(self, callback):
    self.watchers = []
    self.callback = callback

  def updateWatchList(self, watchlist):
    for path in watchlist:
      if not self.isWatching(path):
        self.addWatcher(FileWatcher(path, self.callback))
    for watcher in self.watchers:
      if watcher.path not in watchlist:
        self.removeWatcher(watcher)

  def addWatcher(self, watcher):
    print('[Watchdog] Watching file', watcher.path)
    self.watchers.append(watcher)

  def removeWatcher(self, watcher):
    print('[Watchdog] No longer watching file', watcher.path)
    watcher.stop()
    del self.watchers[self.watchers.index(watcher)]

  def isWatching(self, path):
    return path in map(lambda w: w.path, self.watchers)

  def stopAll(self):
    for watcher in self.watchers:
      watcher.stop()


class FileWatcher:
  def __init__(self, path, callback):
    dir = os.path.dirname(path)
    fn = path.split('/')[-1]
    eventHandler = EventHandler(fn, callback)
    self.path = path
    self.observer = Observer()
    self.observer.schedule(eventHandler, dir, recursive=False)
    self.observer.start()

  def stop(self):
    self.observer.stop()
    self.observer.join()


class EventHandler(PatternMatchingEventHandler):
  def __init__(self, fn, callback, *args):
    super().__init__(patterns=['*/'+fn])
    self.callback = callback
    self.args = args

  def on_modified(self, event):
    self.callback(*self.args)



from Namespace import Namespace

ShpDefinitions = Namespace(
  TagOpen = '[',
  TagClose = ']',
  TagNameScoped = '$',
  TagNameScopeless = '%',
  TagId = '#',
  TagClass = '.',
  TagFlagParam = '!',
  ScopeOpen = '{',
  ScopeClose = '}',
  FunctionName = '@',
  Literal = '\'',
  Comment = '//',
  DefaultDoctype = 'HTML',
)



from shp.shp import SHP

class MultiSHP:
  def __init__(self):
    self.shps = []

  def add(self, source, target):
    self.shps += [SHP(source, target)]

  def compile(self):
    for shp in self.shps:
      shp.compile()

  def watch(self, noBlock=False):
    for shp in self.shps:
      shp.onUpdate()
    if not noBlock: self.run()

  def stop(self):
    for shp in self.shps:
      shp.stop()

  def run(self):
    print('[SHP] Press Ctrl+C to stop')
    try:
      while True:
        sleep(.1)
    except KeyboardInterrupt:
      self.stop()



from Namespace import Namespace
from shp.compiler.LangFuncsDefinitions import definitions

class Call:
  def __init__(self, node):
    self.node = node
    self.func = definitions[self.node.tagName]
    self.ns = self.traverseForNs()

  def __str__(self):
    return f'<Call {self.node.tagName} {self.node.pos}>'

  def execute(self, manager):
    self.func(manager, self)

  def traverseForNs(self):
    ns = []
    node = self.node.parent
    while node is not None:
      if node.tagName == 'namespace':
        ns = [node.params.id] + ns
      node = node.parent
    return ns


class LangFuncsManager:
  def __init__(self, compiler):
    self.compiler = compiler
    self.forcedExecOrder = ['define', 'paste', 'namespace']
    self.reset()

  def reset(self):
    self.defs = []
    self.calls = []

  def executeTree(self, tree):
    self.calls = self.findAllCalls(tree)
    self.sortForcedCalls()
    for call in self.calls:
      call.execute(self)

  def findAllCalls(self, tree):
    calls = []
    for child in tree.children:
      if child.isType('Function'):
        calls.append(Call(child))
      if child.isType('Scoped'):
        calls += self.findAllCalls(child)
    return calls

  def sortForcedCalls(self):
    order = self.forcedExecOrder
    newOrder = []
    for name in order:
      newOrder += list(filter(lambda c: c.node.tagName == name, self.calls))
    for call in self.calls:
      if call not in newOrder: newOrder.append(call)
    self.calls = newOrder



from shp.compiler.RootSource import RootSource
import os

class Source(RootSource):
  def __init__(self, path, parent, dep):
    super().__init__(path)
    self.parent = parent
    self.dep = dep

  @classmethod
  def fromDependency(cls, dep):
    self = cls.__new__(cls)
    directory = os.path.abspath(os.path.dirname(dep.source.path))
    path = directory.replace('\\','/') + f'/{dep.file}.shp'
    super(cls, self).__init__(path)
    self.parent = dep.source
    self.dep = dep
    return self

  def __str__(self):
    return f'<Source "{self.path}">'



from shp.lexer.Position import Position
from shp.lexer.Token import Token
from shp.compiler.ShpDefinitions import ShpDefinitions as SHPDEF

class LexerState:
  def __init__(self, lexer):
    self.lexer = lexer
    self.structural = [SHPDEF.TagOpen, SHPDEF.TagClose,
      SHPDEF.ScopeOpen, SHPDEF.ScopeClose]
    self.whitespace = list(' \r\n\t')

  def switchState(self, other):
    self.lexer.state = other(self.lexer)

  def nextToken(self, pos):
    self.lexer.tokens.append(self.lexer.currentToken)
    self.lexer.currentToken = Token('', pos)
    return self.lexer.currentToken

  def checkPosMatch(self, line, pos, compare):
    if type(compare) == str:
      return line[pos.char:].startswith(compare)
    else:
      return True in [self.checkPosMatch(line, pos, entry)
        for entry in compare]


class LexerStateDefault(LexerState):
  def tokenize(self, line, pos):
    self.lexer.currentToken.initalize(pos)
    if self.checkWhitespace(line, pos): return
    if self.checkIsComment(line, pos): return
    if self.checkStructural(line, pos): return
    self.checkIsLiteral(line, pos)
    self.lexer.currentToken.append(line[pos.char])

  def checkIsComment(self, line, pos):
    if self.checkPosMatch(line, pos, SHPDEF.Comment):
      self.switchState(LexerStateComment)
      return True

  def checkIsLiteral(self, line, pos):
    if self.checkPosMatch(line, pos, SHPDEF.Literal):
      self.switchState(LexerStateLiteral)
      return True

  def checkStructural(self, line, pos):
    token = self.lexer.currentToken
    if self.checkPosMatch(line, pos, self.structural):
      if not token.isEmpty():
        token = self.nextToken(pos)
      token.append(line[pos.char])
      token = self.nextToken(pos)
      return True

  def checkWhitespace(self, line, pos):
    token = self.lexer.currentToken
    if self.checkPosMatch(line, pos, self.whitespace):
      if not token.isEmpty():
        token = self.nextToken(pos)
      return True


class LexerStateComment(LexerState):
  def tokenize(self, line, pos):
    if self.checkPosMatch(line, pos, '\n'):
      self.switchState(LexerStateDefault)


class LexerStateLiteral(LexerState):
  def tokenize(self, line, pos):
    token = self.lexer.currentToken
    token.append(line[pos.char])
    if self.checkPosMatch(line, pos, SHPDEF.Literal):
      self.nextToken(pos)
      self.switchState(LexerStateDefault)




class BuilderState:
  def __init__(self, builder):
    self.builder = builder

  def switchState(self, state):
    self.builder.state = state(self.builder)

  def buildParameters(self, node):
    if not len(node.params.keys()): return ''
    result = ''
    for key, val in node.params.items():
      if key in node.paramLiterals:
        val = f"'{val}'"
      result += f' {key}={val}'
    return result


class BuilderStateScopeless(BuilderState):
  def build(self, node):
    if self.checkIsData(node): return
    if self.checkIsScoped(node): return
    self.builder.html += f'<{node.tagName}{self.buildParameters(node)}>'

  def checkIsData(self, node):
    if node.tagName == '__data__':
      self.switchState(BuilderStateData)
      self.builder.state.build(node)
      return True

  def checkIsScoped(self, node):
    if node.isType('Scoped'):
      self.switchState(BuilderStateScoped)
      self.builder.state.build(node)
      return True


class BuilderStateScoped(BuilderStateScopeless):
  def build(self, node):
    if self.checkIsData(node): return
    if self.checkIsScopeless(node): return
    self.builder.html += f'<{node.tagName}{self.buildParameters(node)}>'
    for child in node.children:
      self.build(child)
    self.builder.html += f'</{node.tagName}>'

  def checkIsScopeless(self, node):
    if not node.isType('Scoped'):
      self.switchState(BuilderStateScopeless)
      self.builder.state.build(node)
      return True


class BuilderStateData(BuilderState):
  def build(self, node):
    if self.checkIsNonData(node): return
    self.builder.html += node.data

  def checkIsNonData(self, node):
    if node.tagName != '__data__':
      self.switchState(BuilderStateScoped)
      self.builder.state.build(node)
      return True



import os
from argparse import ArgumentParser
from Namespace import Namespace
from shp import SHP

def main(args):
  source = os.path.abspath(args.source).replace('\\', '/')
  target = os.path.abspath(args.target).replace('\\', '/')
  shp = SHP(source, target)
  if args.watch:
    shp.watch()
  else:
    shp.compile()

if __name__ == '__main__':
  ap = ArgumentParser()
  ap.add_argument('source', help='Path to the SHP source file')
  ap.add_argument('target', help='Path to the target HTML file')
  ap.add_argument('-w', '--watch',
    help='Use this flag to recompile whenever the source file is edited',
    action='store_true')
  main(ap.parse_args())



from shp.builder.BuilderStates import BuilderStateScoped

class Builder:
  def __init__(self):
    self.state = BuilderStateScoped(self)
    self.html = ''

  def build(self, tree):
    for node in tree.children:
      self.state.build(node)
    return self.html



from shp.compiler.Compiler import Compiler
from shp.Watchdog import Watchdog
from time import sleep

class SHP:
  def __init__(self, source, target):
    self.compiler = Compiler(source)
    self.watchdog = Watchdog(lambda: self.onUpdate())
    self.source = source
    self.target = target

  def compile(self):
    print('[SHP] Compiling', self.target)
    html = self.compiler.compile()
    with open(self.target, 'w') as file:
      file.write(html)

  def watch(self, noBlock=False):
    self.onUpdate()
    if not noBlock: self.run()

  def onUpdate(self):
    self.compile()
    self.watchdog.updateWatchList([self.source] + self.compiler.includedFiles)

  def stop(self):
    print('[SHP] Interrupted')
    self.watchdog.stopAll()

  def run(self):
    print('[SHP] Press Ctrl+C to stop')
    try:
      while True:
        sleep(.1)
    except KeyboardInterrupt:
      self.stop()



from shp.lexer.Lexer import Lexer
from shp.parser.Parser import Parser
from shp.parser.Nodes import NodeFunction
from shp.compiler.Dependency import Dependency

def sanitizePath(path):
  return path.replace('\\', '/')


class RootSource:
  def __init__(self, path):
    self.path = path

  def parse(self):
    self.root = self.parseFile(self.path)
    self.deps = self.traverseForDeps(self.root)

  def __str__(self):
    return f'<Source "{self.path}" (root)>'

  def parseFile(self, path):
    path = sanitizePath(path)
    with open(path, 'r') as file:
      tokens = Lexer().tokenize(file.read())
      return Parser().parse(tokens)

  def resolve(self, subSources):
    for source in subSources:
      node = source.dep.includeNode
      newNode = NodeFunction('namespace', node.pos)
      newNode.params.id = node.params['as']
      newNode.children = source.root.children
      for child in newNode.children:
        child.parent = newNode
        child.updateDepth(node.depth+1)
      node.replaceSelf(newNode)

  def traverseForDeps(self, node):
    deps = []
    if node.isType('Function') and node.tagName == 'include':
      deps.append(Dependency(self, node))
    elif node.isType('Scoped'):
      for child in node.children:
        deps += self.traverseForDeps(child)
    return deps



from Namespace import Namespace
from shp.compiler.Definition import Definition
from shp.compiler.ShpDefinitions import ShpDefinitions as SHPDEF
from shp.parser.Nodes import NodeData

def _define(manager, call):
  manager.defs.append(Definition(call))
  call.node.removeSelf()

def _paste(manager, call):
  id = call.node.params.id
  ns = call.ns[:]
  try: ns += [call.node.params['from']]
  except KeyError: pass
  ns = '/'.join(ns)
  matching = list(filter(lambda d: d.id == id and d.ns == ns, manager.defs))
  if len(matching) > 1:
    raise ValueError(str(call)+' More than 1 definition is found')
  if len(matching) == 0:
    raise ValueError(str(call)+' No definitions were found')
  definition = matching[0]
  call.node.replaceSelfList(definition.children)

def _namespace(manager, call):
  if len(call.node.children):
    call.node.replaceSelfList(call.node.children)
  else:
    call.node.removeSelf()

def _doctype(manager, call):
  try: value = call.node.params.id
  except AttributeError: value = SHPDEF.DefaultDoctype
  node = NodeData(f'<!DOCTYPE {value}>', call.node.pos)
  call.node.replaceSelf(node)


definitions = Namespace(
  define = _define,
  paste = _paste,
  namespace = _namespace,
  doctype = _doctype,
)



from shp.compiler.RootSource import RootSource
from shp.compiler.Source import Source
from shp.compiler.LangFuncsManager import LangFuncsManager
from shp.builder.Builder import Builder
import os

class Compiler:
  def __init__(self, entryPoint):
    self.root = RootSource(entryPoint)
    self.funcs = LangFuncsManager(self)
    self.reset()

  def reset(self):
    self.funcs.reset()
    self.includedFiles = []

  def compile(self):
    self.reset()
    self.root.parse()
    self._compile(self.root)
    tree = self.root.root
    tree.updateDepth(-1)
    self.funcs.executeTree(tree)
    return Builder().build(tree)

  def _compile(self, source):
    subSources = self.resolveDeps(source)
    for sub in subSources:
      self.includedFiles += [sub.path]
      sub.parse()
      self._compile(sub)
    source.resolve(subSources)

  def resolveDeps(self, source):
    sources = []
    for dep in source.deps:
      sources.append(Source.fromDependency(dep))
    return sources



from shp.parser.ParserStates import ParserStateDefault
from shp.parser.Nodes import NodeScoped
from shp.lexer.Position import Position

class Parser:
  def __init__(self):
    self.reset()

  def reset(self):
    self.state = ParserStateDefault(self)
    self.root = NodeScoped('Root', Position(0,0))
    self.root.depth = -1
    self.currentScope = self.root
    self.lastAddedTag = self.root

  def parse(self, tokens):
    self.reset()
    for token in tokens:
      self.state.parse(token)
    return self.root



from shp.compiler.ShpDefinitions import ShpDefinitions

class Dependency:
  def __init__(self, source, includeNode):
    self.source = source
    self.file = self.sanitize(includeNode.params.file)
    self.includeNode = includeNode

  def __str__(self):
    return f'<Dependency "{self.file}" from "{self.source.path}" {self.includeNode.pos}>'

  @staticmethod
  def sanitize(path):
    return path.replace(' ', '')\
      .replace(ShpDefinitions.Literal, '')



from shp.parser.Nodes import Node, NodeScoped, NodeFunction
from Namespace import Namespace

class ParserState:
  def __init__(self, parser):
    self.parser = parser
    self.tagPrefixes = ['TagNameScoped', 'TagNameScopeless', 'FunctionName']
    self.paramPrefixes = ['TagId', 'TagClass', 'TagFlagParam']

  def switchState(self, other):
    self.parser.state = other(self.parser)

  def checkType(self, token, compare, depth=0):
    if type(compare) == str:
      return token.type == compare
    else:
      return True in [self.checkType(token, entry, depth+1)
        for entry in compare]


class ParserStateDefault(ParserState):
  def parse(self, token):
    if self.checkTagOpen(token): return
    if self.checkTagName(token): return
    if self.checkScopeChange(token): return
    self.parser.currentScope.appendData(token.data, token.pos)

  def checkTagOpen(self, token):
    if self.checkType(token, 'TagOpen'):
      self.switchState(ParserStateTag)
      return True

  def checkTagName(self, token):
    if self.checkType(token, self.tagPrefixes):
      node = Namespace(
        TagNameScopeless = Node,
        TagNameScoped = NodeScoped,
        FunctionName = NodeFunction
      )[token.type](token.data[1:], token.pos)
      self.parser.currentScope.appendNode(node)
      self.parser.lastAddedTag = node
      return True

  def checkScopeChange(self, token):
    if self.checkType(token, ['ScopeOpen', 'ScopeClose']):
      self.switchState(ParserStateScope)
      self.parser.state.parse(token)
      return True


class ParserStateTag(ParserState):
  def __init__(self, parser):
    super().__init__(parser)
    self.index = 0;
    self.lastKey = '';

  def parse(self, token):
    if self.checkType(token, 'TagClose'):
      self.switchState(ParserStateDefault)
      return
    if self.checkType(token, self.paramPrefixes):
      self.parsePrefixedParam(token)
      return
    self.parseIndexedParam(token)

  def parsePrefixedParam(self, token):
    node = self.parser.lastAddedTag
    value = token.data[1:]
    Namespace(
      TagId = lambda n: n.addParam('id', value),
      TagClass = lambda n: n.addParam('class', value),
      TagFlagParam = lambda n: n.addParam(value, True),
    )[token.type](node)

  def parseIndexedParam(self, token):
    self.index += 1
    if self.index % 2:
      self.lastKey = token.data
    else:
      self.parser.lastAddedTag.addParam(self.lastKey, token.data)


class ParserStateScope(ParserState):
  def parse(self, token):
    if self.checkType(token, 'ScopeOpen'):
      self.parser.currentScope = self.parser.lastAddedTag
    elif self.checkType(token, 'ScopeClose'):
      self.parser.currentScope = self.parser.currentScope.parent
    self.switchState(ParserStateDefault)



from shp.compiler.ShpDefinitions import ShpDefinitions as SHPDEF
from Namespace import Namespace

class Node:
  def __init__(self, tagName, pos):
    self.tagName = tagName
    self.pos = pos
    self.scoped = False
    self.params = Namespace()
    self.paramLiterals = []
    self.types = ['Node']
    # set when added to the tree
    self.parent = None
    self.depth = 0

  def __str__(self):
    indent = self.depth * '  '
    return f'{indent}{self.shortStr()}\n'

  def shortStr(self):
    return f'<{self.tagName} {self.pos}/>'

  def addParam(self, key, val):
    if not key in self.params.keys():
      self.params[key] = val
    else:
      self.params[key] += ' ' + val
      if key not in self.paramLiterals:
        self.paramLiterals.append(key)

  def isType(self, type):
    return type in self.types

  def updateDepth(self, depth):
    self.depth = depth


class NodeData(Node):
  def __init__(self, data, pos):
    super().__init__('__data__', pos)
    self.types.append('Data')
    self.data = data

  def __str__(self):
    indent = self.depth * '  '
    return f'{indent}{self.shortStr()}\n'

  def shortStr(self):
    return f'"{self.data}" {self.pos}'

  def appendData(self, data):
    self.data += (' ' if len(self.data) else '') + data


class NodeScoped(Node):
  def __init__(self, tagName, pos):
    super().__init__(tagName, pos)
    self.types.append('Scoped')
    self.children = []

  def __str__(self):
    indent = self.depth * '  '
    header = f'{indent}{self.shortStr()}'
    if not len(self.children):
      return f'{header}\n'
    result = f'{header} {{\n'
    for child in self.children:
      result += str(child)
    return result + f'{indent}}}\n'

  def shortStr(self):
    return f'<{self.tagName} {self.pos}>'

  def appendNode(self, node):
    self.children.append(node)
    node.parent = self
    node.depth = self.depth + 1

  def appendData(self, data, pos):
    try: self._appendDataToLastChild(data)
    except IndexError:
      lastChild = NodeData('', pos)
      self.appendNode(lastChild)
      lastChild.appendData(data)

  def _appendDataToLastChild(self, data):
    lastChild = self.children[-1]
    if not lastChild.isType('Data'): raise IndexError
    lastChild.appendData(data)

  def updateDepth(self, depth):
    self.depth = depth
    for child in self.children:
      child.updateDepth(depth+1)


class NodeFunction(NodeScoped):
  def __init__(self, tagName, pos):
    super().__init__(tagName, pos)
    self.types.append('Function')

  def shortStr(self):
    return f'@({self.tagName} {self.pos})'

  def __str__(self):
    indent = self.depth * '  '
    header = f'{indent}{self.shortStr()}'
    if not len(self.children):
      return f'{header}\n'
    result = f'{header} {{\n'
    for child in self.children:
      result += str(child)
    return result + f'{indent}}}\n'

  def removeSelf(self):
    index = self.parent.children.index(self)
    del self.parent.children[index]

  def replaceSelf(self, node):
    index = self.parent.children.index(self)
    self.parent.children[index] = node
    node.parent = self.parent
    node.updateDepth(self.depth)

  def replaceSelfList(self, nodes):
    index = self.parent.children.index(self)
    del self.parent.children[index]
    self.parent.children[index:index] = nodes
    for child in nodes:
      child.parent = self.parent
      child.updateDepth(self.depth)




class Definition:
  def __init__(self, call):
    self.id = call.node.params.id
    self.ns = '/'.join(call.ns)
    self.children = call.node.children

  def __str__(self):
    if len(self.ns):
      return f'<Definition {self.id} in "{self.ns}">'
    else:
      return f'<Definition {self.id}>'



from shp.lexer.Position import Position
from shp.compiler.ShpDefinitions import ShpDefinitions as SHPDEF

class Token:
  def __init__(self, data='', pos=None):
    self.data = data
    self.pos = Position(0,0) if pos is None else pos
    self.detectType()

  def __str__(self):
    return f'<Token {self.type} "{self.data}" {self.pos}>'

  def initalize(self, pos):
    if self.isEmpty():
      self.pos = pos

  def append(self, data):
    self.data += data
    self.detectType()

  def isEmpty(self):
    return self.data == ''

  def detectType(self):
    updated = False
    for key, prefix in SHPDEF.items():
      if self.data.startswith(prefix):
        self.type = key
        updated = True
    if not updated:
      self.type = 'Text'



from shp.lexer.Position import Position
from shp.lexer.Token import Token
from shp.lexer.LexerStates import LexerStateDefault

class Lexer:
  def __init__(self):
    self.reset()

  def reset(self):
    self.tokens = []
    self.currentToken = Token()
    self.state = LexerStateDefault(self)

  def tokenize(self, shp):
    self.reset()
    lines = shp.split('\n')
    for lineNo, line in enumerate(lines):
      line += '\n' # for comment end detection
      for charNo in range(len(line)):
        self.state.tokenize(line, Position(lineNo, charNo))
    return self.tokens



from setuptools import setup, find_packages

setup(
  name = 'shp',
  description = 'Static-Html-Preprocessor',
  version = '1.0.2',
  license = 'MIT',
  author = 'Jakub21',
  url = 'https://github.com/Jakub21/shp.py',
  packages = find_packages(),
)
